# import Converter
# import ModuleLoader
# import utils
from .Converter import convert
# from .utils import *
from .utils import extract_function_calls
from .utils import isVerilogLine
from .utils import isModuleFunc
from .utils import findPythonVarinVerilogLine
from .utils import processVerilogLine
from .utils import processPythonVarinVerilogInst
from .utils import processVerilogLine_str
from .utils import parseVerilog_inst_block
from .utils import processVerlog_inst_line
from .utils import processVerilog_inst_block
from .utils import judge_state
from .utils import state_transition
from .utils import extract_vparam_ports
from .utils import instantiate_full
from .utils import instantiate
from .utils import replace_single_quotes
from .ModuleLoader import ModuleLoader
from .ModuleLoader import moduleloader
from .ModuleLoader import ModuleLoader_Singleton


# from pytv.utils import *
from functools import wraps
import inspect
import re
from functools import wraps
from inspect import getcallargs
import warnings
import json
import hashlib
from pytv.utils import extract_function_calls
from pytv.utils import isVerilogLine
from pytv.utils import isModuleFunc
from pytv.utils import findPythonVarinVerilogLine
from pytv.utils import processVerilogLine
from pytv.utils import processPythonVarinVerilogInst
from pytv.utils import processVerilogLine_str
from pytv.utils import parseVerilog_inst_block
from pytv.utils import processVerlog_inst_line
from pytv.utils import processVerilog_inst_block
from pytv.utils import judge_state
from pytv.utils import state_transition
from pytv.utils import extract_vparam_ports
from pytv.utils import instantiate_full
from pytv.utils import instantiate
from pytv.utils import replace_single_quotes
import pytv.ModuleLoader
from pytv.ModuleLoader import ModuleLoader
from pytv.ModuleLoader import ModuleLoader_Singleton
from pytv.ModuleLoader import moduleloader
# the decorator replaces func with a newly defined function decorated(*args, **kwargs)
def convert(func):
    def decorated(*args, **kwargs):
        STATE = 'IN_PYTHON'
        func_name = func.__name__
        abstract_module_name = func_name[6:]
        concrete_module_name = abstract_module_name
        # Add decorated func to the list in moduleloader
        moduleloader.add_module_func(func_name)
        flag_end_inst = 0
        new_func_code = []
        inst_code = []
        # the generated verilog code.
        python_vars_dict = kwargs
        src_lines, starting_line = inspect.getsourcelines(func)
        i = 0
        # definition of the newly generated function for producing v_declaration code
        line_func_def = f"def func_new(*args, **kwargs): \n"
        for line in src_lines:
            i = i + 1
            if i == 3:
                # pass the keyword variables
                for key in kwargs.keys():
                    new_func_code.append(f"    {key}=kwargs['{key}']\n")
                new_func_code.append(f"    abstract_module_name = '{abstract_module_name}'\n")
                new_func_code.append(' ' * 4 + 'v_declaration = str()\n')
                new_func_code.append(' ' * 4 + 'v_module_name_tree = dict()\n')
                new_func_code.append(' ' * 4 + 'v_module_dict_list = []\n')
            stripped_line = line.strip()
            if stripped_line.startswith("return"):
                tokens = stripped_line.split()
                if tokens[0] == "return":
                    continue
            STATE = judge_state(stripped_line)
            if STATE == 'IN_PYTHON':
                new_func_code.append(line + '\n')
            elif STATE == 'IN_VERILOG_INLINE':
                line_renew = processVerilogLine(stripped_line)
                line_renew_str = processVerilogLine_str(stripped_line)
                new_func_code.append(
                    ' ' * (len(line) - len(stripped_line) - 1) + line_renew_str + '\n')
            elif STATE == 'BEGIN_VERILOG_INST':
                flag_end_inst = 1
                inst_code.append(line + '\n')
            elif STATE == 'IN_VERILOG_INST':
                if not isVerilogLine(line):
                    b = isVerilogLine(line)
                    inst_line_renew = processVerlog_inst_line(line)
                    #Here the following 2 lines are added
                    #v_declaration_in, module_dict_tree_in, module_file_name_in = ModuleBasic(p1=10, p2=10)
                    #v_module_dict_list.append(module_dict_tree_in)
                    new_func_code.append(inst_line_renew)
                else:
                    inst_code.append(line + '\n')
                inst_code = []
            elif STATE == 'END_VERILOG_INST':
                inst_code = []
        new_func_code.pop(0)
        new_func_code.pop(0)
        new_func_code.insert(0, line_func_def)
        new_func_code.append("    return v_declaration, v_module_dict_list")
        new_func_body = ''.join(new_func_code[0:])
        local_vars = {}
        exec(new_func_body, func.__globals__, local_vars)  # 执行新函数代码
        func_new = local_vars['func_new']
        verilog_code, module_dict_list = func_new(*args, **kwargs)
        module_dict_tree = dict()
        module_generated, module_file_name, inst_idx_str = moduleloader.generate_module(abstract_module_name, python_vars_dict, verilog_code)
        inst_verilog_code, module_name_tmp = instantiate_full(verilog_code, kwargs, module_file_name, inst_idx_str)
        module_dict_tree[module_file_name] = module_dict_list
        moduleloader.generate_file_tree(module_dict_tree)
        # pass the instantiation information to the singleton module
        moduleloader.add_module_inst_info(inst_verilog_code, verilog_code, module_dict_tree, concrete_module_name, func_name)
    return decorated




# setup.py
from setuptools import setup, find_packages

setup(
    name='verithon_test',
    version='0.2',
    packages=find_packages(),
    description='RTL Auto-generation with Verilog Embedded in Python',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='JiaYanXu',
    author_email='jiayanxu@seu.edu.cn',
    url='https://github.com/autohdw/Voldelog.git',
    license='MIT',
    python_requires='>=3.6',
)



from functools import wraps
import inspect
import re
from functools import wraps
from inspect import getcallargs
import warnings
import json
import hashlib
from pytv.ModuleLoader import moduleloader
from pytv.ModuleLoader import ModuleLoader_Singleton
from pytv.ModuleLoader import ModuleLoader


# Warning COLOR Setings
RED = "\033[31m"
RESET = "\033[0m"

def extract_function_calls(line):
    pattern = r'\b(\w+)\s*\('
    matches = re.findall(pattern, line)
    return matches

def isVerilogLine(line):
    line = line.strip()
    pattern = '^#/'
    return bool(re.match(pattern, line))

def isModuleFunc(line):
    is_mf = False
    # for mf_name in ModuleLoader_Singleton.module_func_list:
    #     if mf_name in line:
    #         if not "def" in line:
    #             is_mf = True
    #             return is_mf
    line = line.strip()
    func_names = []
    func_name = str()
    if "def" in line:
        return is_mf
    if line.startswith('#'):
        return is_mf
    func_names = extract_function_calls(line)
    if len(func_names) > 0:
        func_name = func_names[0]
        if func_name.startswith("Module"):
            is_mf = True
            return is_mf
    return is_mf

def findPythonVarinVerilogLine(line):
    is_found = False
    pattern = r'(?<=\`).+?(?=\`)'
    var_names = []
    var_names = re.findall(pattern, line)
    if len(var_names) > 0:
        is_found = True
    return var_names, is_found

# def subPythonVarinVerilog(line, local_var_dict):
#     [python_var_names, has_python_var] = findPythonVarinVerilogLine(line)
#     for python_var_name in python_var_names

def processVerilogLine(line):
    [python_var_names, has_python_var] = findPythonVarinVerilogLine(line)

    # Converting the python variable to numbers with python function str()
    # This is done when the decorator fucntion is executed, here only the python code is generated
    if(has_python_var):
        for python_var_name in python_var_names:
            name_str_extended = '`'+ python_var_name + '`'
            # This seems robust because there are usually no numbers at the start or end of a verilog line
            replace_str = "'+" + "str(" + python_var_name + ")" + "+'"
            line = line.replace(name_str_extended, replace_str)
    idx = line.index('#/')
    line_cut = line[idx+2:]
    line_cut_extend = "'" + line_cut + "'"
    line_code_renew = "print(" + line_cut_extend + ")"
    return line_code_renew

# directly replace the
def processPythonVarinVerilogInst(line, python_var_dict):
    [python_var_names, has_python_var] = findPythonVarinVerilogLine(line)
    if has_python_var:
        for python_var_name in python_var_names:
            name_str_extended = '`' + python_var_name + '`'
            replace_str = str(python_var_dict.get(python_var_name))
            line = line.replace(name_str_extended, replace_str)
    return line

def processVerilogLine_str(line):
    [python_var_names, has_python_var] = findPythonVarinVerilogLine(line)
    # fix for bugs in lines such as a = 12'b`llr`
    line = replace_single_quotes(line, "\\'")
    # Converting the python variable to numbers with python function str()
    # This is done when the decorator fucntion is executed, here only the python code is generated
    if(has_python_var):
        for python_var_name in python_var_names:
            name_str_extended = '`'+ python_var_name + '`'
            # This seems robust because there are usually no numbers at the start or end of a verilog line
            replace_str = "'+" + "str(" + python_var_name + ")" + "+'"
            line = line.replace(name_str_extended, replace_str)
    idx = line.index('#/')
    line_cut = line[idx+2:]
    #line_cut_extend = "'" + line_cut +  + "'"
    line_cut_extend = f"'{line_cut}\\n'"
    line_code_renew = 'v_declaration'+'='+'v_declaration'+'+'+line_cut_extend
    return line_code_renew

# This function processes the verilog instance block and returns a line of python code
def parseVerilog_inst_block(kwargs, module_file_name_in, inst_idx_str):
    has_module_name = False
    has_inst_name = False
    has_vparams = False
    #inst_keys = kwargs.keys()
    PORT_DICT = {}
    VPARAM_DICT = {}
    PARAM_DICT = []
    INST_NAME = str()
    MODULE_NAME = str()
    isTOP = True
    if 'PORTS' in kwargs:
        PORT_DICT = kwargs['PORTS']
        isTOP = False
    elif not moduleloader.disable_warning:
        warnings.warn(f"{RED}Call of module function with unassigned ports is detected. Make sure this is the top module{RESET}",stacklevel=4)
    if 'VPARAMS' in kwargs:
        has_vparams = True
        VPARAM_DICT = kwargs['VPARAMS']

    if 'MODULE_NAME' in kwargs:
        has_module_name = True
        MODULE_NAME = kwargs['MODULE_NAME']
        #if len(kwargs['MODULE_NAME'] > 0):
            
    if 'INST_NAME' in kwargs:
        has_inst_name = True
        INST_NAME = kwargs['INST_NAME']
        # if len(kwargs['INST_NAME'] > 0):
        
    # Assigning default module name if not defined in the verilog instance block
    module_file_name_in = module_file_name_in.strip()
    module_file_name_in = module_file_name_in.replace(".v","")
    module_file_name_in = module_file_name_in.strip()
    if not has_module_name:
        if not moduleloader.disable_warning:
           warnings.warn(f"{RED}Module name is not specified in the verilog instance block. Default module name applied.{RESET}",stacklevel=4)
        MODULE_NAME = module_file_name_in
    # Assigning default inst name if not defined in the verilog instance block
    if not has_inst_name:
        if not moduleloader.disable_warning:
           warnings.warn(f"{RED}Inst name is not specified in the verilog instance block. Default Inst name applied.{RESET}",stacklevel=4)
        INST_NAME = 'u_' + inst_idx_str + '_'
        INST_NAME = INST_NAME + MODULE_NAME
    # func_name = 'Module'+ MODULE_NAME
    return PORT_DICT, PARAM_DICT, VPARAM_DICT, INST_NAME, MODULE_NAME, isTOP

# process the python function for instantiating a verilog module by adding the return value
def processVerlog_inst_line(inst_line):
    # print(inst_line)
    inst_line_strip = inst_line.strip()
    n_blanks = len(inst_line) - len(inst_line_strip)
    # print(n_blanks)
    inst_line_renew0 = " " * (n_blanks-1) + inst_line_strip + "\n"
    inst_line_renew1 = " " * (n_blanks-1) + 'v_inst_code_in, v_declaration_in, module_dict_tree_in, module_file_name_in = '+ 'moduleloader.extract_module_inst_info()' + "\n"
    inst_line_renew2 = " " * (n_blanks-1) + f"v_module_dict_list.append(module_dict_tree_in) \n"
    inst_line_renew3 = " " * (n_blanks-1) + f"v_declaration = v_declaration + v_inst_code_in \n"
    inst_line_renew = inst_line_renew0 + inst_line_renew1 + inst_line_renew2 + inst_line_renew3
    return inst_line_renew


def processVerilog_inst_block(inst_code):
    inst_code_str = str(inst_code)
    line_renew = f"inst_v_code_tmp, module_name_tmp = instantiate_full(v_declaration_in,{inst_code_str}, locals(), kwargs, module_file_name_in) \n"
    return line_renew

def judge_state(line):
    if isVerilogLine(line):
        STATE = 'IN_VERILOG_INLINE'
    else:
        if isModuleFunc(line):
            STATE = 'IN_VERILOG_INST'
        else:
            STATE = 'IN_PYTHON'
    return STATE

def state_transition(STATE_prev, line):
    match STATE_prev:
        case 'IN_PYTHON':
            if isVerilogLine(line):
                line_without_note = line.replace('#/', '')
                line_strip = line_without_note.strip()
                if line_strip.startswith('INST:'):
                    STATE = 'BEGIN_VERILOG_INST'
                else:
                    STATE = 'IN_VERILOG_INLINE'
            else:
                STATE = 'IN_PYTHON'
        
        case 'IN_VERILOG_INLINE':
            if isVerilogLine(line):
                line_without_note = line.replace('#/', '')
                line_strip = line_without_note.strip()
                if line_strip.startswith('INST:'):
                    STATE = 'BEGIN_VERILOG_INST'
                else:
                    STATE = 'IN_VERILOG_INLINE'
            else:
                STATE = 'IN_PYTHON'

        case 'BEGIN_VERILOG_INST':
            # TEST
            # print("INST BEGINS! \n")
            if isVerilogLine(line):
                line_without_note = line.replace('#/', '')
                line_strip = line_without_note.strip()
                if line_strip.startswith('INST:'):
                    STATE = 'BEGIN_VERILOG_INST'
                    raise Exception("Error: Nested verilog instance block is not supported in the current PyTv version.")
                elif line_strip.startswith('ENDINST'):
                    STATE = 'END_VERILOG_INST'
                else:
                    STATE = 'IN_VERILOG_INST'
            else:
                STATE = 'IN_VERILOG_INST'
                #raise Exception("Error: No python code inside the verilog instance block.")
     

        case 'IN_VERILOG_INST':
            if isVerilogLine(line):
                line_without_note = line.replace('#/', '')
                line_strip = line_without_note.strip()
                if line_strip.startswith('INST:'):
                    STATE = 'BEGIN_VERILOG_INST'
                    raise Exception("Error: Nested verilog instance block is not supported in the current PyTv version.")
                elif line_strip.startswith('ENDINST'):
                    STATE = 'END_VERILOG_INST'
                else:
                    STATE = 'IN_VERILOG_INST'
            else:
                STATE = 'IN_VERILOG_INST'
                #raise Exception("Error: No python code inside the verilog instance block.")

            
        case 'END_VERILOG_INST':
            if isVerilogLine(line):
                line_without_note = line.replace('#/', '')
                line_strip = line_without_note.strip()
                if line_strip.startswith('INST:'):
                    STATE = 'BEGIN_VERILOG_INST'
                else:
                    STATE = 'IN_VERILOG_INLINE'
            else:
                STATE = 'IN_PYTHON'
    return STATE

def extract_vparam_ports(v_declaration):
    pattern = r'(?<=\().+?(?=\))'
    v_declaration = v_declaration.replace('\n',' ')
    port_names = str()
    vparam_names = []
    vparam_and_port_names = re.findall(pattern, v_declaration)
    if not vparam_and_port_names:
        vparam_names = []
        port_names = []
        return vparam_names, port_names
    port_names = vparam_and_port_names[0].strip()
    if not port_names:
        vparam_names = []
        port_names = []
        return vparam_names, port_names
    segments = port_names.split(',')
    last_words = [segment.strip().split()[-1] for segment in segments if segment.strip()]
    #port_names = vparam_and_port_names[1].replace(',',' ')
    #vparam_names = vparam_names.split()
    port_names = last_words
    return vparam_names, port_names


def instantiate_full(v_declaration, kwargs, module_file_name_in,inst_idx_str):
    PORT_DICT_real = dict()
    VPARAM_DICT_real = dict()
    # Find the required ports by looking up the verilog code
    [vparams_names, ports_names] = extract_vparam_ports(v_declaration)
    # Get the ports and vparams passed by kwargs
    [PORT_DICT, PARAM_DICT, VPARAM_DICT, INST_NAME, MODULE_NAME, isTOP] = parseVerilog_inst_block(kwargs, module_file_name_in, inst_idx_str)
    v_code = str()
    # Exit function if it represents the top module
    if (isTOP):
        return v_code, MODULE_NAME
    if not isinstance(PORT_DICT,dict):
        if not isinstance(PORT_DICT,list):
            raise Exception(f"Module Instantiation: PORTS argument must receive a dict or list")
            pass
    if (len(vparams_names) < len(VPARAM_DICT)) or (len(ports_names) < len(PORT_DICT)):
        raise Exception(f"Module Instantiation: Not enough ports to connect. Expected {len(PORT_DICT)}, got {len(ports_names)}")
        pass
    if isinstance(PORT_DICT, dict):
        PORT_DICT_real = PORT_DICT
    elif isinstance(PORT_DICT, list):
        cnt = 0
        for port_name in ports_names:
            if cnt == len(PORT_DICT):
                break
            PORT_DICT_real[port_name] = PORT_DICT[cnt]
            cnt = cnt + 1
    v_code = instantiate(PORT_DICT_real, VPARAM_DICT_real, INST_NAME, MODULE_NAME)
    return v_code, MODULE_NAME

def instantiate(ports_dict, vparams_dict, module_name, inst_name):
        params_str = str()
        ports_str = str()
        verilog_code = str()
        # parameters
        if len(ports_dict) > 0:
            ports_str = ", ".join([f".{key}({value})" for key, value in ports_dict.items()])
        # generate verilog code for instantiation
        verilog_code = f"{inst_name}  {module_name}({ports_str});\n"
        # print verilog code for instantiation
        return verilog_code

def replace_single_quotes(input_string, replacement):
    result = ''
    in_backticks = False
    i = 0

    while i < len(input_string):
        if input_string[i] == '`':
            in_backticks = not in_backticks

        if input_string[i] == "'" and not in_backticks:
            result += replacement
        else:
            result += input_string[i]

        i += 1

    return result